/*****************************************************************************/
/*!
 * \file expr_value.h
 * 
 * Author: Sergey Berezin
 * 
 * Created: Fri Feb  7 15:07:18 2003
 *
 * <hr>
 * Copyright (C) 2003 by the Board of Trustees of Leland Stanford
 * Junior University and by New York University. 
 *
 * License to use, copy, modify, sell and/or distribute this software
 * and its documentation for any purpose is hereby granted without
 * royalty, subject to the terms and conditions defined in the \ref
 * LICENSE file provided with this distribution.  In particular:
 *
 * - The above copyright notice and this permission notice must appear
 * in all copies of the software and related documentation.
 *
 * - THE SOFTWARE IS PROVIDED "AS-IS", WITHOUT ANY WARRANTIES,
 * EXPRESSED OR IMPLIED.  USE IT AT YOUR OWN RISK.
 * 
 * <hr>
 * 
 * Class ExprValue: the value holding class of Expr.  No one should
 * use it directly; use Expr API instead.  To enforce that, the
 * constructors are made protected, and only Expr, ExprManager, and
 * subclasses can use them.
 */
/*****************************************************************************/

// *** HACK ATTACK *** (trick from Aaron Stump's code)

// In order to inline the Expr constructors (for efficiency), this
// file (expr_value.h) must be included in expr.h.  However, we also
// need to include expr.h here, hence, circular dependency.  A way to
// break it is to include expr_value.h in the middle of expr.h after
// the definition of class Expr, but before the definition of its
// inlined methods.  So, expr.h included below will also suck in
// expr_value.h recursively, meaning that we then should skip the rest
// of the file (since it's already been included). 

// That's why expr.h is outside of #ifndef.  The same is true for
// type.h and theorem.h.

#ifndef _CVC_lite__expr_h_
#include "expr.h"
#endif

#ifndef _CVC_lite__expr_value_h_
#define _CVC_lite__expr_value_h_

// To define NULL
#include <stdlib.h>

#include "theorem.h"
#include "type.h"
#include "debug.h"
#include "compat_hash_set.h"
#include "cdo.h"
#include "memory_manager.h"

// The prime number used in the hash function for a vector of elements
#define PRIME 131

namespace CVCL {
  
  class Expr;
  class Expr::iterator;
  class ExprManager;


/*****************************************************************************/
/*!
 *\class ExprValue
 *\brief The base class for holding the actual data in expressions
 * 
 *
 * Author: Sergey Berezin
 *
 * Created: long time ago
 *
 * \anchor ExprValue The base class just holds the operator.
 * All the additional data resides in subclasses.
 * 
 */
/*****************************************************************************/
class ExprValue {
    friend class Expr;
    friend class Expr::iterator;
    friend class ExprManager;
    friend class CInterface;
  private:
    // Index for compare() method.  This makes the ordering be the
    // creation order, which is useful for some DPs
    ExprIndex d_index;
    // Reference counter for garbage collection
    int d_refcount;
    //! Generic flag for marking expressions (e.g. in DAG traversal)
    unsigned d_flag;


    // Commonly used attributes

  protected:
    // Static hash functions.  They don't depend on the context
    // (ExprManager and such), so it is still thread-safe to have them
    // static.
    static std::hash<char*> s_charHash;
    static std::hash<long int> s_intHash;

    static size_t pointerHash(void* p) { return s_intHash((long int)p); }
    // Hash function for subclasses with children
    static size_t hash(const int kind, const std::vector<Expr>& kids);
    // Hash function for kinds
    static size_t hash(const int n) { return s_intHash((long int)n); }

    // Our expr. manager
    ExprManager* d_em;
    // The kind of the expression.  In particular, it determines which
    // subclass of ExprValue is used to store the expression.
    int d_kind;

    //! Non-caching hash function which actually computes the hash.
    /*! This is the method that all subclasses should implement */
    virtual size_t computeHash() const { return hash(d_kind); }
    //! Rebuild expression with a new ExprManager (no uniquifying)
    /*! A subclass has to redefine this method whenever it has
     * additional Expr fields (kids, operator, etc.).  See
     * ExprNode::rebuild() for an example. */
    virtual ExprValue* rebuild(ExprManager* newEM) const
     { return copy(newEM); }
    //! Recursive call to rebuild() [do NOT redefine!]
    /*! Use in subclasses to rebuild additional expressions */
    Expr rebuildExpr(ExprManager* newEM, const Expr& e) const {
      return newEM->rebuildRec(e);
    }
    //! Rebuild all kids at once [do NOT redefine!]
    /*! Use in subclasses to rebuild additional expressions */
    void rebuildExprs(ExprManager* newEM, const std::vector<Expr>& kids,
		      std::vector<Expr>& newKids) const {
      newEM->rebuild(kids, newKids);
    }

  private:

    size_t d_hash; //!< Cached hash value (initially 0)

    void incRefcount() { ++d_refcount; }

    void decRefcount() {
      // Cannot be DebugAssert, since we are called in a destructor
      // and should not throw an exception
      ASSERT_FATAL(d_refcount > 0, "Mis-handled the ref. counting");
      if((--d_refcount) == 0) d_em->gc(this);
    }

    //! Set the ExprIndex (to avoid recreating an ExprValue)
    void setIndex(ExprIndex idx) { d_index = idx; }

    // Theorem and Type can be Null; in this case, it corresponds to
    // having no such attribute.
    //! The find attribute (may be NULL)
    CDO<Theorem>* d_find;
    //! The cached type of the expression (may be Null)
    Type d_type;
    //! The cached TCC of the expression (may be Null)
    Expr d_tcc;
    //! Subtyping predicate for the expression and all subexpressions
    Theorem d_subtypePred;

    // Notify list may be NULL (== no such attribute)
    NotifyList* d_notifyList;

    // For caching calls to Simplify
    Theorem d_simpCache;
    unsigned d_simpCacheTag;

    // bit-vector used for flags that are not context-dependent
    unsigned d_staticFlags;

    // The base class has no children
    virtual int arity() const { return 0; }

    //! Caching hash function
    /*! Do NOT implement it in subclasses! Implement computeHash() instead.
     */
    size_t hash() const {
      if(d_hash == 0)
	const_cast<ExprValue*>(this)->d_hash = computeHash();
      return d_hash;
    }

    // Check if this (sub)class has a non-trivial operator
//     virtual bool hasOp() const { return false; }
    // Special attributes for uninterpreted functions
    virtual CDO<Theorem>* getSig() const {
      DebugAssert(false, "getSig() is called on ExprValue");
      return NULL;
    }
    virtual CDO<Theorem>* getRep() const {
      DebugAssert(false, "getRep() is called on ExprValue");
      return NULL;
    }
    virtual void setSig(CDO<Theorem>* sig) {
      DebugAssert(false, "setSig() is called on ExprValue");
    }
    virtual void setRep(CDO<Theorem>* rep) {
      DebugAssert(false, "setRep() is called on ExprValue");
    }

    int d_height;
    int getHeight() const { return d_height; }

    int d_highestKid;
    int getHighestKid() const { return d_highestKid; }

    Expr d_simpFrom;
    const Expr& getSimpFrom() const { return d_simpFrom; }
    void setSimpFrom(const Expr& simpFrom) { d_simpFrom = simpFrom; }

  protected:
    //! Return the memory manager (for the benefit of subclasses)
    MemoryManager* getMM(size_t MMIndex) {
      DebugAssert(d_em!=NULL, "ExprValue::getMM()");
      return d_em->getMM(MMIndex);
    }

  public:
    //! Test whether the expression is a generic subclass
    /*!
     * \return 0 for the core classes, and getMMIndex() value for
     * generic subclasses (those defined in DPs)
     */
    virtual size_t isGeneric() const { return 0; }
    /*! @brief Test whether the expression is a generic subclass with
     * the given MM index
     */ 
    virtual bool isGeneric(size_t idx) const { return false; }
    //! String expression tester
    virtual bool isString() const { return false; }
    //! Rational number expression tester
    virtual bool isRational() const { return false; }
    //! Uninterpreted constants
    virtual bool isVar() const { return false; }
    //! Bound variables
    virtual bool isBoundVar() const { return false; }
    //! Boolean variable
    virtual bool isBoolVar() const { return false; }
    //! Application of a LAMBDA-expression
    virtual bool isApply() const { return false; }
    //! A LAMBDA-expression or a quantifier
    virtual bool isClosure() const { return false; }

  //! Constructor
    ExprValue(ExprManager* em, int kind, ExprIndex idx = 0)
      : d_index(idx), d_refcount(0), d_flag(0),
	d_em(em), d_kind(kind), d_hash(0), d_find(NULL), d_notifyList(NULL),
	d_simpCacheTag(0), //d_simpInPlace(NULL),
      d_staticFlags(0), d_height(0)
    {
      DebugAssert(em != NULL, "NULL ExprManager is given to ExprValue()");
      DebugAssert(em->isKindRegistered(kind),
		  ("ExprValue(kind = " + int2string(kind)
		   + ")): kind is not registered").c_str());
    }
    //! Destructor
    virtual ~ExprValue();

    //! Equality between any two ExprValue objects (including subclasses)
    virtual bool operator==(const ExprValue& ev2) const;
    //! Make a clean copy of itself using the given memory manager
    virtual ExprValue* copy(ExprManager* em, ExprIndex idx = 0) const;
    //! Make a clean copy of itself with new children and given memory manager
    /*! This version of copy() is used for creating an expression from
     * operator (class Op) and kids.
     */
    virtual ExprValue* copy(ExprManager* em, const std::vector<Expr>& kids,
			    ExprIndex idx = 0) const;

    //! Tell ExprManager who we are
    virtual size_t getMMIndex() const { return EXPR_VALUE; }
    int getKind() const { return d_kind; }

    // Return a ref to an empty vector (stored in ExprManager)
    virtual const std::vector<Expr>& getKids() const;
    // Methods to access leaf data in subclasses
    virtual const std::string& getUid() const { 
      static std::string null;
      DebugAssert(false, "ExprValue::getUid() called in base class");
      return null;
    }
    virtual const std::string& getString() const {
      DebugAssert(false, "getString() is called on ExprValue");
      static std::string s("");
      return s;
    }

    virtual const Rational& getRational() const {
      DebugAssert(false, "getRational() is called on ExprValue");
      static Rational r(0);
      return r;
    }

//     virtual const Expr& getOpRef() const;
//     virtual Expr getOp() const { return getOpRef(); }

    //! Returns the string name of UCONST and BOUND_VAR expr's.
    virtual const std::string& getName() const;
    //! Returns the original Boolean variable (for BoolVarExprValue)
    virtual const Expr& getVar() const {
      DebugAssert(false, "getVar() is called on ExprValue");
      static Expr null;
      return null;
    }
  //! Get the function Expr from a function application
    virtual const Expr& getFun() const {
      DebugAssert(false, "getFun() is called on ExprValue");
      static Expr null;
      return null;
    }

    virtual const std::vector<Expr>& getVars() const  {
      DebugAssert(false, "getVars() is called on ExprValue");
      static std::vector<Expr> null;
      return null;
    }

    virtual const Expr& getBody() const {
      DebugAssert(false, "getBody() is called on ExprValue");
      static Expr null;
      return null;
    }

    virtual const Expr& getExistential() const {
      DebugAssert(false, "getExistential() is called on ExprValue");
      static Expr null;
      return null;
    }
    virtual int getBoundIndex() const {
      DebugAssert(false, "getIndex() is called on ExprValue");
      return 0;
    }

    virtual const std::vector<std::string>& getFields() const {
      DebugAssert(false, "getFields() is called on ExprValue");
      static std::vector<std::string> null;
      return null;
    }
    virtual const std::string& getField() const {
      DebugAssert(false, "getField() is called on ExprValue");
      static std::string null;
      return null;
    }
    virtual int getTupleIndex() const {
      DebugAssert(false, "getTupleIndex() is called on ExprValue");
      return 0;
    }

    // Generic values (parts of the expression)

    virtual int getIntValue(int idx) const;
    virtual const std::string& getStringValue(int idx) const;
    virtual const Rational& getRatValue(int idx) const;
    virtual const Expr& getExprValue(int idx) const;
    virtual const Theorem& getTheoremValue(int idx) const;
    virtual const CDO<int>& getCDIntValue(int idx) const;
    virtual const CDO<Expr>& getCDExprValue(int idx) const;
    virtual const CDO<Theorem>& getCDTheoremValue(int idx) const;
    // Values of all other types (don't use it unless absolutely needed)
    virtual const void* getOtherValue(int idx) const;

    // Get the number of the corresponding generic values
    virtual size_t getIntValueSize() const { return 0; }
    virtual size_t getStringValueSize() const { return 0; }
    virtual size_t getRatValueSize() const { return 0; }
    virtual size_t getExprValueSize() const { return 0; }
    virtual size_t getTheoremValueSize() const { return 0; }
    virtual size_t getCDIntValueSize() const { return 0; }
    virtual size_t getCDExprValueSize() const { return 0; }
    virtual size_t getCDTheoremValueSize() const { return 0; }
    // Values of all other types (don't use it unless absolutely needed)
    virtual size_t getOtherValueSize() const { return 0; }

    // Generic attributes
    /*! \name Generic Attributes
     * \anchor ExprValueGenericAttr 
     *
     * Functions returning an assignable reference to generic
     * attributes.  Assigning to the reference returned by each
     * function will set the appropriate attribute.
     *
     * The attributes of each type are indexed by the idx parameter.
     *
     * @{
     */

    virtual int& getIntAttr(int idx);
    virtual Expr& getExprAttr(int idx);
    virtual Theorem& getTheoremAttr(int idx);
    virtual CDO<int>& getCDIntAttr(int idx);
    virtual CDO<Expr>& getCDExprAttr(int idx);
    virtual CDO<Theorem>& getCDTheoremAttr(int idx);
    //! Attributes of all other types (don't use it unless absolutely needed)
    virtual void* getOtherAttr(int idx);
    /*! @} */

    // Get the number of the corresponding generic attributes
    virtual size_t getIntAttrSize() const { return 0; }
    virtual size_t getExprAttrSize() const { return 0; }
    virtual size_t getTheoremAttrSize() const { return 0; }
    virtual size_t getCDIntAttrSize() const { return 0; }
    virtual size_t getCDExprAttrSize() const { return 0; }
    virtual size_t getCDTheoremAttrSize() const { return 0; }
    // Attributes of all other types (don't use it unless absolutely needed)
    virtual size_t getOtherAttrSize() const { return 0; }

    // Memory management
    void* operator new(size_t size, MemoryManager* mm) {
      return mm->newData(size);
    }
    void operator delete(void*) { }

  }; // end of class ExprValue

  // Class ExprNode; it's an expression with children
  class ExprNode: public ExprValue {
    friend class Expr;
    friend class ExprManager;
  private:
    // Tell ExprManager who we are
    size_t getMMIndex() const { return EXPR_NODE; }
  protected:
    std::vector<Expr> d_children;

    int arity() const { return d_children.size(); }

    const std::vector<Expr>& getKids() const { return d_children; }
   // Use our static hash() for the member method
    size_t computeHash() const {
      return ExprValue::hash(d_kind, d_children);
    }
    //! Rebuild expression with a new ExprManager (no uniquifying)
    ExprValue* rebuild(ExprManager* newEM) const {
      std::vector<Expr> newKids;
      rebuildExprs(newEM, d_children, newKids);
      return copy(newEM, newKids);
    }

  public:
    //! Constructors.  Kids are assumed to be non-empty.
    ExprNode(ExprManager* em, int kind, const std::vector<Expr>& kids,
	     unsigned idx = 0)
      : ExprValue(em, kind, idx), d_children(kids) {
    }
    //! Destructor
    virtual ~ExprNode() { }
    
    virtual bool operator==(const ExprValue& ev2) const;
    //! Make a clean copy of itself using the given memory manager
    virtual ExprValue* copy(ExprManager* em, ExprIndex idx = 0) const;
    virtual ExprValue* copy(ExprManager* em, const std::vector<Expr>& kids,
			    ExprIndex idx = 0) const;
    // Memory management
    void* operator new(size_t size, MemoryManager* mm) {
      return mm->newData(size);
    }
    void operator delete(void*) { }
  }; // end of class ExprNode

  // Leaf expressions
  class ExprString: public ExprValue {
    friend class Expr;
    friend class ExprManager;
  private:
    std::string d_str;

    // Hash function for this subclass
    static size_t hash(const std::string& str) {
      return s_charHash(str.c_str());
    }

    // This Expr has no kids
//     virtual const std::vector<Expr>& getKids() const;

    // Tell ExprManager who we are
    virtual size_t getMMIndex() const { return EXPR_STRING; }

  protected:
    // Use our static hash() for the member method
    virtual size_t computeHash() const { return hash(d_str); }

    virtual bool isString() const { return true; }
    virtual const std::string& getString() const { return d_str; }

  public:
    // Constructor
    ExprString(ExprManager* em, const std::string& s, unsigned idx = 0)
      : ExprValue(em, STRING_EXPR, idx), d_str(s) { }
    // Destructor
    virtual ~ExprString() { }

    virtual bool operator==(const ExprValue& ev2) const;
    //! Make a clean copy of itself using the given memory manager
    virtual ExprValue* copy(ExprManager* em, ExprIndex idx = 0) const;
    // Memory management
    void* operator new(size_t size, MemoryManager* mm) {
      return mm->newData(size);
    }
    void operator delete(void*) { }
  }; // end of class ExprString


  class ExprSkolem: public ExprValue {
    friend class Expr;
    friend class ExprManager;
  private:
    Expr d_quant; //!< The quantified expression to skolemize
    int d_idx; //!< Variable index in the quantified expression
    const Expr& getExistential() const {return d_quant;}
    int getBoundIndex() const {return d_idx;}

    // virtual const std::vector<Expr>& getKids() const;
    
    // Tell ExprManager who we are
    size_t getMMIndex() const { return EXPR_SKOLEM;}

  protected:
    size_t computeHash() const {
      size_t res = getExistential().getBody().hash();
      res = PRIME*res + getBoundIndex();
      return res;
    }

    bool operator==(const ExprValue& ev2) const;

    //! Make a clean copy of itself using the given memory manager
    ExprValue* copy(ExprManager* em, ExprIndex idx = 0) const;
    ExprValue* rebuild(ExprManager* newEM) const;
    bool isVar() const { return true; }
   
  public:
    // Constructor
    ExprSkolem(ExprManager* em, int index, const Expr& exist, unsigned idx = 0)
      : ExprValue(em, SKOLEM_VAR, idx), d_quant(exist), d_idx(index) { }
    // Destructor
    virtual ~ExprSkolem() { }
    // Memory management
    void* operator new(size_t size, MemoryManager* mm) {
      return mm->newData(size);
    }
    void operator delete(void*) { }
  }; // end of class ExpSkolem

  class ExprRational: public ExprValue {
    friend class Expr;
    friend class ExprManager;
  private:
    Rational d_r;

    virtual const Rational& getRational() const { return d_r; }

    // Hash function for this subclass
    static size_t hash(const Rational& r) {
      return s_charHash(r.toString().c_str());
    }

    // This Expr has no kids
    virtual const std::vector<Expr>& getKids() const;

    // Tell ExprManager who we are
    virtual size_t getMMIndex() const { return EXPR_RATIONAL; }

  protected:

    virtual size_t computeHash() const { return hash(d_r); }
    virtual bool operator==(const ExprValue& ev2) const;
    //! Make a clean copy of itself using the given memory manager
    virtual ExprValue* copy(ExprManager* em, ExprIndex idx = 0) const;
    virtual bool isRational() const { return true; }

  public:
    // Constructor
    ExprRational(ExprManager* em, const Rational& r, unsigned idx = 0)
      : ExprValue(em, RATIONAL_EXPR, idx), d_r(r) { }
    // Destructor
    virtual ~ExprRational() { }
    // Memory management
    void* operator new(size_t size, MemoryManager* mm) {
      return mm->newData(size);
    }
    void operator delete(void*) { }
  }; // end of class ExprRational

  // Uninterpreted constants (variables)
  class ExprVar: public ExprValue {
    friend class Expr;
    friend class ExprManager;
  private:
    std::string d_name;

    virtual const std::string& getName() const { return d_name; }

    // This Expr has no kids
    virtual const std::vector<Expr>& getKids() const;

    // Tell ExprManager who we are
    virtual size_t getMMIndex() const { return EXPR_UCONST; }
  protected:

    virtual size_t computeHash() const {
      return s_charHash(d_name.c_str());
    }
    virtual bool isVar() const { return true; }

  public:
    // Constructor
    ExprVar(ExprManager *em, const std::string& name, unsigned idx = 0)
      : ExprValue(em, UCONST, idx), d_name(name) { }
    // Destructor
    virtual ~ExprVar() { }

    virtual bool operator==(const ExprValue& ev2) const;
    //! Make a clean copy of itself using the given memory manager
    virtual ExprValue* copy(ExprManager* em, ExprIndex idx = 0) const;
    // Memory management
    void* operator new(size_t size, MemoryManager* mm) {
      return mm->newData(size);
    }
    void operator delete(void*) { }
  }; // end of class ExprVar



  class ExprBoundVar: public ExprValue {
    friend class Expr;
    friend class ExprManager;
  private:
    std::string d_name;
    std::string d_uid;

    virtual const std::string& getName() const { return d_name; }
    virtual const std::string& getUid() const { return d_uid; }

    // This Expr has no kids
    virtual const std::vector<Expr>& getKids() const;

    // Tell ExprManager who we are
    virtual size_t getMMIndex() const { return EXPR_BOUND_VAR; }
  protected:

    virtual size_t computeHash() const {
      return s_charHash(d_name.c_str())*PRIME + s_charHash(d_uid.c_str());
    }
    virtual bool isVar() const { return true; }
    virtual bool isBoundVar() const { return true; }

  public:
    // Constructor
    ExprBoundVar(ExprManager *em, const std::string& name,
		 const std::string& uid, unsigned idx = 0)
      : ExprValue(em, BOUND_VAR, idx), d_name(name), d_uid(uid) { }
    // Destructor
    virtual ~ExprBoundVar() { }

    virtual bool operator==(const ExprValue& ev2) const;
    //! Make a clean copy of itself using the given memory manager
    virtual ExprValue* copy(ExprManager* em, ExprIndex idx = 0) const;
    // Memory management
    void* operator new(size_t size, MemoryManager* mm) {
      return mm->newData(size);
    }
    void operator delete(void*) { }
  }; // end of class ExprBoundVar

  // Application of an uninterpreted function
  class ExprCC: public ExprNode {
    friend class Expr;
    friend class ExprManager;
  protected:
    // Special attributes for this type of expression
    CDO<Theorem>* d_sig;
    CDO<Theorem>* d_rep;
    // computeHash() is the same as in ExprNode
  public:
    // Constructor
    ExprCC(ExprManager *em, int kind,
	   const std::vector<Expr>& kids, unsigned idx = 0)
      : ExprNode(em, kind, kids, idx), d_sig(NULL), d_rep(NULL) { }
    // Destructor
    virtual ~ExprCC();

    virtual bool operator==(const ExprValue& ev2) const;
    //! Make a clean copy of itself using the given memory manager
    virtual ExprValue* copy(ExprManager* em, ExprIndex idx = 0) const;
    virtual ExprValue* copy(ExprManager* em, const std::vector<Expr>& kids,
			    ExprIndex idx = 0) const;
    // Tell ExprManager who we are
    virtual size_t getMMIndex() const { return EXPR_CC; }

    virtual CDO<Theorem>* getSig() const { return d_sig; }
    virtual CDO<Theorem>* getRep() const { return d_rep; }

    virtual void setRep(CDO<Theorem>* rep) { d_rep = rep; }
    virtual void setSig(CDO<Theorem>* sig) { d_sig = sig; }
    // Memory management
    void* operator new(size_t size, MemoryManager* mm) {
      return mm->newData(size);
    }
    void operator delete(void*) { }
  }; // end of class ExprCC


  class ExprApply: public ExprCC {
    friend class Expr;
    friend class ExprManager;
  private:
    Expr d_fun;
  protected:
    size_t getMMIndex() const { return EXPR_APPLY; }
    size_t computeHash() const {
      return PRIME*ExprCC::computeHash() + d_fun.hash();
    }
    const Expr& getFun() const { return d_fun; }
    bool isApply() const { return true; }
    // Make a clean copy of itself using the given memory manager
    ExprValue* copy(ExprManager* em, ExprIndex idx = 0) const;
    ExprValue* copy(ExprManager* em, const std::vector<Expr>& kids,
		    ExprIndex idx = 0) const;
    ExprValue* rebuild(ExprManager* em) const;
  public:
    // Constructor
    ExprApply(ExprManager* em, const Expr& func,
	      const std::vector<Expr>& kids, unsigned idx = 0)
      : ExprCC(em, APPLY, kids, idx), d_fun(func) { }
    virtual ~ExprApply() { }
    bool operator==(const ExprValue& ev2) const;
    // Memory management
    void* operator new(size_t size, MemoryManager* mm) {
      return mm->newData(size);
    }
    void operator delete(void*) { }
  }; // end of class ExprApply


  /*! @brief A "closure" expression which binds variables used in the
      "body".  Used by LAMBDA and quantifiers. */
  class ExprClosure: public ExprValue {
    friend class Expr;
    friend class ExprManager;
  private:
    //! Bound variables
    std::vector<Expr> d_vars;
    //! The body of the quantifier/lambda
    Expr d_body;

    //! Tell ExprManager who we are
    virtual size_t getMMIndex() const { return EXPR_CLOSURE; }

    virtual const std::vector<Expr>& getVars() const { return d_vars; }
    virtual const Expr& getBody() const { return d_body; }


  protected:

    size_t computeHash() const;

  public:
    // Constructor
    ExprClosure(ExprManager *em, int kind, const std::vector<Expr>& vars,
	      const Expr& body, unsigned idx = 0)
      : ExprValue(em, kind, idx), d_vars(vars), d_body(body) { }
    // Destructor
    virtual ~ExprClosure() { }

    bool operator==(const ExprValue& ev2) const;
    //! Make a clean copy of itself using the given memory manager
    ExprValue* copy(ExprManager* em, ExprIndex idx = 0) const;
    ExprValue* rebuild(ExprManager* newEM) const;
    // Memory management
    void* operator new(size_t size, MemoryManager* mm) {
      return mm->newData(size);
    }
    void operator delete(void*) { }
    virtual bool isClosure() const { return true; }
  }; // end of class ExprClosure


/*****************************************************************************/
/*!
 *\class NamedExprValue
 *\brief NamedExprValue
 *
 * Author: Clark Barrett
 *
 * Created: Thu Dec  2 23:18:17 2004
 *
 * Subclass of ExprValue for kinds that have a name associated with them.
 */
/*****************************************************************************/

  class NamedExprValue : public ExprNode {
    friend class Expr;
    friend class ExprManager;
  private:
    std::string d_name; //! name of constructor/selector/tester

  protected:

    ExprValue* copy(ExprManager* em, ExprIndex idx = 0) const {
      return new(em->getMM(getMMIndex()))
	NamedExprValue(d_em, d_kind, d_name, d_children, idx);
    }

    ExprValue* copy(ExprManager* em, const std::vector<Expr>& kids,
		    ExprIndex idx = 0) const {
      return new(em->getMM(getMMIndex()))
	NamedExprValue(d_em, d_kind, d_name, kids, idx);
    }

    size_t computeHash() const {
      return s_charHash(d_name.c_str())*PRIME + ExprNode::computeHash();
    }

    size_t getMMIndex() const { return EXPR_NAMED; }
  
  public:
    // Constructor
    NamedExprValue(ExprManager *em, int kind, const std::string& name,
                   const std::vector<Expr>& kids, ExprIndex idx = 0)
      : ExprNode(em, kind, kids, idx), d_name(name) { }
    // virtual ~NamedExprValue();
    bool operator==(const ExprValue& ev2) const {
      if(getMMIndex() != ev2.getMMIndex()) return false;
      return (getName() == ev2.getName())
	&& ExprNode::operator==(ev2);
    }

    const std::string& getName() const { return d_name; }

    // Memory management
    void* operator new(size_t size, MemoryManager* mm) {
      return mm->newData(size);
    }
    void operator delete(void*) { }
  }; // end of class NamedExprValue

  //! Boolean variables are treated as 0-ary predicates
  class BoolVarExprValue: public ExprValue {
    friend class Expr;
    friend class ExprManager;
  private:
    Expr d_op;
  protected:
    // Methods
    size_t computeHash() const { return d_op.hash() + PRIME; }
    size_t getMMIndex() const { return EXPR_BOOL_VAR; }
  public:
    // Constructor
    BoolVarExprValue(ExprManager* em, const Expr& op, unsigned idx = 0)
      : ExprValue(em, BOOL_VAR, idx), d_op(op) { }

    bool isBoolVar() const { return true; }
    const Expr& getVar() const { return d_op; }

    bool operator==(const ExprValue& ev2) const {
      return (ev2.isBoolVar() && d_op == ev2.getVar());
    }
    ExprValue* copy(ExprManager* em, ExprIndex idx = 0) const {
      return new(em->getMM(getMMIndex())) BoolVarExprValue(d_em, d_op, idx);
    }
    ExprValue* rebuild(ExprManager* em) const {
      return new(em->getMM(getMMIndex()))
	BoolVarExprValue(d_em, rebuildExpr(em, d_op));
    }
    // Memory management
    void* operator new(size_t size, MemoryManager* mm) {
      return mm->newData(size);
    }
    void operator delete(void*) { }
  };

} // end of namespace CVCL

#endif
