/*****************************************************************************/
/*!
 * \file search.cpp
 * 
 * Author: Clark Barrett
 * 
 * Created: Fri Jan 17 14:19:54 2003
 *
 * <hr>
 * Copyright (C) 2003 by the Board of Trustees of Leland Stanford
 * Junior University and by New York University. 
 *
 * License to use, copy, modify, sell and/or distribute this software
 * and its documentation for any purpose is hereby granted without
 * royalty, subject to the terms and conditions defined in the \ref
 * LICENSE file provided with this distribution.  In particular:
 *
 * - The above copyright notice and this permission notice must appear
 * in all copies of the software and related documentation.
 *
 * - THE SOFTWARE IS PROVIDED "AS-IS", WITHOUT ANY WARRANTIES,
 * EXPRESSED OR IMPLIED.  USE IT AT YOUR OWN RISK.
 * 
 * <hr>
 * 
 */
/*****************************************************************************/

#include "search.h"
#include "context.h"
#include "theory.h"
#include "eval_exception.h"
#include "search_rules.h"
#include "variable.h"

using namespace CVCL;
using namespace std;

// class SearchEngine::Splitter methods

// Constructor
SearchEngine::Splitter::Splitter(const Literal& lit): d_lit(lit) {
d_lit.count()++;
  TRACE("Splitter", "Splitter(", d_lit, ")");
}

// Copy constructor
SearchEngine::Splitter::Splitter(const SearchEngine::Splitter& s)
  : d_lit(s.d_lit) {
  d_lit.count()++;
  TRACE("Splitter", "Splitter[copy](", d_lit, ")");
}

// Assignment
SearchEngine::Splitter&
SearchEngine::Splitter::operator=(const SearchEngine::Splitter& s) {
  if(this == &s) return *this; // Self-assignment
  d_lit.count()--;
  d_lit = s.d_lit;
  d_lit.count()++;
  TRACE("Splitter", "Splitter[assign](", d_lit, ")");
  return *this;
}

// Destructor
SearchEngine::Splitter::~Splitter() {
  d_lit.count()--;
  TRACE("Splitter", "~Splitter(", d_lit, ")");
}



//! Constructor
SearchEngine::SearchEngine(VCL* vcl)
  : d_vcl(vcl), d_core(vcl->theoryCore()),
    d_dpSplitters(vcl->getCurrentContext()),
    d_notifyObj(this, vcl->getCurrentContext()),
    d_bottomScope(vcl->getCurrentContext(), 0),
    d_assumptions(vcl->getCurrentContext()),
    d_skolemized_thms(vcl->getCurrentContext()),
    d_skolemVars(vcl->getCurrentContext()),
    d_cnfCache(vcl->getCurrentContext()),
    d_cnfVars(vcl->getCurrentContext()),
    d_cnfOption(&(vcl->getFlags()["cnf"].getBool())),
    d_ifLiftOption(&(vcl->getFlags()["iflift"].getBool())),
    d_ignoreCnfVarsOption(&(vcl->getFlags()["ignore-cnf-vars"].getBool())),
    d_origFormulaOption(&(vcl->getFlags()["orig-formula"].getBool())),
    d_trackLiterals(&(vcl->getFlags()["track-literals"].getBool())),
    d_enqueueCNFCache(vcl->getCurrentContext()),
    d_applyCNFRulesCache(vcl->getCurrentContext()),
    d_replaceITECache(vcl->getCurrentContext())
{
  d_rules = createRules();
  d_vm = new VariableManager(vcl->getCM(), d_rules,
			     vcl->getFlags()["mm"].getString());
  IF_DEBUG(d_assumptions.setName("CDList[SearchEngine::d_assumptions]"));
}


//! Destructor
SearchEngine::~SearchEngine()
{
  delete d_rules;
  delete d_vm;
}


// Returns a new theorem if e has not already been asserted, otherwise returns
// a NULL theorem.
Theorem SearchEngine::newUserAssertion(const Expr& e) {
  Theorem thm;
  CDMap<Expr,Theorem>::iterator i(d_assumptions.find(e));
  d_literals.clear();
  IF_DEBUG(if(debugger.trace("search verbose")) {
    ostream& os(debugger.getOS());
    os << "d_assumptions = [";
    for(CDMap<Expr,Theorem>::iterator i=d_assumptions.begin(),
	  iend=d_assumptions.end(); i!=iend; ++i) {
      debugger.getOS() << "(" << (*i).first << " => " << (*i).second << "), ";
    }
    os << "]" << endl;
  });
  if(i!=d_assumptions.end()) {
    TRACE("search verbose", "newUserAssertion(", thm,
	  "): assumption already exists");
  } else {
    thm = d_core->assumpRule(e);
    d_assumptions[e] = thm;
    TRACE("search verbose", "newUserAssertion(", thm,
	  "): created new assumption");
  }
  return thm;
}

Theorem SearchEngine::skolemize(const Theorem& thm, bool addToModel) { 
  const Expr& e = thm.getExpr();
  if(e.isExists()) {
    TRACE("skolem", "Skolemizing existential:", "", "{");
    CDMap<Expr, Theorem>::iterator i=d_skolemized_thms.find(e),
      iend=d_skolemized_thms.end();
    if(i!=iend) {
      TRACE("skolem", "Skolemized theorem found in map: ", (*i).second, "}");
      return d_core->iffMP(thm, (*i).second);
    }
    
    Theorem skol = d_rules->skolemizeRewrite(e);
    for(unsigned int i=0; i<e.getVars().size(); i++) {
      Expr skolV(e.skolemExpr(i));
      Type tp(d_core->getType(e.getVars()[i]));
      skolV.setType(tp);
      if(tp.isBool()) {
	skolV = newBoolVarExpr(skolV);
	skolV.setType(tp);
      }
    }

    d_query_axioms.push_back(skol);
    d_skolemized_thms.insert(e, skol, d_bottomScope);
    skol = d_core->iffMP(thm, skol);
    TRACE("skolem", "skolemized new theorem: ", skol, "}");
    return skol;
  }
  return thm;
}

/*!
 * This is equivalent to skolemize(d_core->varIntroRule(e)), only more
 * efficient.
 */
Theorem SearchEngine::varIntroSkolem(const Expr& e, bool addToModel) {
  // First, look up the cache
  CDMap<Expr, Theorem>::iterator i=d_skolemVars.find(e),
    iend=d_skolemVars.end();
  if(i!=iend) return (*i).second;
  // Not in cache; create a new one
  Theorem thm = d_core->varIntroRule(e);
  const Expr& e2 = thm.getExpr();
  DebugAssert(e2.isExists() && e2.getVars().size()==1, "varIntroSkolem: e2 = "
	      +e2.toString());
  Theorem skolThm;
  // Check if we have a corresponding skolemized version already
  CDMap<Expr, Theorem>::iterator j=d_skolemized_thms.find(e2),
    jend=d_skolemized_thms.end();
  if(j!=jend) {
    skolThm = (*i).second;
  } else {
    skolThm = d_rules->skolemizeRewriteVar(e2);
    d_query_axioms.push_back(skolThm);
    d_skolemized_thms.insert(e2, skolThm, d_bottomScope);
    // Add the new Skolem var to the model, if needed
    if(addToModel) {
      const Expr& skolEq = skolThm.getRHS();
      DebugAssert(skolEq.arity()==2 && skolEq[1].isSkolem(),
		  "skolEq = "+skolEq.toString());
      Expr skolV(skolEq[1]);
      d_core->addToVarDB(skolV);
    }
  }
  thm = d_core->iffMP(thm, skolThm);
  d_skolemVars.insert(e, thm, d_bottomScope);
  return thm;
}


// Same as newUserAssertion, except it's an error to assert something that's
// already been asserted.
Theorem SearchEngine::newAssertion(const Expr& e) {
  Theorem thm;
  DebugAssert(!d_assumptions.count(e),
	      "newAssertion: repeated assertion: "+e.toString());
    thm = d_core->assumpRule(e);
    d_assumptions[e] = thm;
    TRACE("search verbose", "newAssertion(", thm,
	  "): created new assumption");
  return thm;
}


void SearchEngine::getAssertions(std::vector<Expr>& assertions)
{
  for(CDMap<Expr,Theorem>::orderedIterator i=d_assumptions.orderedBegin(),
	iend=d_assumptions.orderedEnd(); i!=iend; ++i)
    assertions.push_back((*i).first);
}


bool SearchEngine::wasAsserted(const Expr& e) {
  return d_assumptions.count(e) > 0;
}


void SearchEngine::addCNFFact(const Theorem& thm, bool fromCore) {
  TRACE("addCNFFact", "addCNFFact(", thm.getExpr(), ") {");
  if(isLiteral(thm.getExpr())) {
    if (*d_trackLiterals) {
      d_literals.push_back(thm.getExpr());
    }
    addLiteralFact(thm);
    // These literals are derived, and  should also be reported to the core.  
    if(!fromCore) d_core->enqueueFact(thm);
  } else {
    addNonLiteralFact(thm);
  }
  TRACE_MSG("addCNFFact", "addCNFFact => }");
}


void SearchEngine::addFact(const Theorem& thm) {
  TRACE("search addFact", "SearchEngine::addFact(", thm.getExpr(), ") {");
  if(*d_cnfOption)
    enqueueCNF(thm);
  else
    addCNFFact(thm, true);
  TRACE_MSG("search addFact", "SearchEngine::addFact => }");
}


void SearchEngine::addSplitter(const Expr& e, int priority) {
  DebugAssert(isLiteral(e), "SearchEngine::addSplitter("+e.toString()+")");
  d_dpSplitters.push_back(Splitter(newLiteral(e)));
}


void SearchEngine::getCounterExample(vector<Expr>& assertions, bool inOrder)
{
  if(!d_vcl->getTM()->withAssumptions())
    throw EvalException
      ("Method getCounterExample() (or command ASSERTIONS) cannot be used\n"
       " without assumptions activated");
  if(!d_lastValid.isNull())
    throw EvalException
      ("Method getCounterExample() (or command ASSERTIONS)\n"
       " must be called only after failed QUERY");
  if(!d_lastCounterExample.empty()) {
    if (inOrder) {
      for(CDMap<Expr,Theorem>::orderedIterator i=d_assumptions.orderedBegin(),
            iend=d_assumptions.orderedEnd(); i!=iend; ++i) {
        if (d_lastCounterExample.find((*i).first) != d_lastCounterExample.end()) {
          assertions.push_back((*i).first);
        }
      }
      DebugAssert(assertions.size()==d_lastCounterExample.size(),
                  "getCounterExample: missing assertion");
    }
    else {
      ExprHashMap<bool>::iterator i=d_lastCounterExample.begin(),
        iend = d_lastCounterExample.end();
      for(; i!=iend; ++i) assertions.push_back((*i).first);
    }
  }
}


void SearchEngine::getConcreteModel(ExprMap<Expr> & m)
{
  TRACE("model" ,  "Building a concrete model", "", "{");
  if(!d_lastValid.isNull())
    throw EvalException
      ("Method getConcreteModel() (or command COUNTEREXAMPLE)\n"
       " must be called only after failed QUERY");
  // Save the scope level, to recover on errors
  int scope = d_vcl->getCurrentContext()->level();
  d_vcl->getCurrentContext()->push();
  d_core->collectBasicVars();
  d_core->assertDisequalities();
  bool v;
  // FIXME: redundant (though probably cheap) call to checkValid
//   v = checkValid(d_vcl->falseExpr());
//   if(v) {
//     d_vcl->getCurrentContext()->popto(scope);
//     throw EvalException
//       ("Model Creation failed after asserting disequalities.\n"
//        "You might be using an incomplete fragment of the theory");
//   }
  try {
    d_core->refineCounterExample();
  } catch(Exception& e) {
    // Clean up and re-throw the exception
    d_vcl->getCurrentContext()->popto(scope);
    throw e;
  }
  v =checkValid(d_vcl->falseExpr());
  if(v) {
    vector<Expr> assump;
    getAssertions(assump);
    d_core->inconsistentThm().getAxiomAssumptions(assump);
    Expr a = d_vcl->listExpr(assump);
    d_vcl->getCurrentContext()->popto(scope);
    throw EvalException
      ("Model Creation failed after refining counterexample\n"
       "due to the following assumptions:\n "
       +a.toString()       
       +"\n\nYou might be using an incomplete fragment of the theory");
  }
  try {
    d_core->buildModel(m);
  } catch(Exception& e) {
    // Clean up and re-throw the exception
    d_vcl->getCurrentContext()->popto(scope);
    throw e;
  }
  v =checkValid(d_vcl->falseExpr());
  if(v) {
    vector<Expr> assump;
    getAssertions(assump);
    Expr a = d_vcl->listExpr(assump);
    d_vcl->getCurrentContext()->popto(scope);
    throw EvalException
      ("Model Creation failed due to the following assumptions:\n"
       +a.toString()
       +"\n\nYou might be using an incomplete fragment of the theory");
  }
  TRACE("model" ,  "Building a concrete model", "", "}"); 
}


Proof 
SearchEngine::getProof()
{
  if(!d_vcl->getTM()->withProof())
    throw EvalException
      ("DUMP_PROOF cannot be used without proofs activated");
  if(d_lastValid.isNull())
    throw EvalException
      ("DUMP_PROOF must be called only after successful QUERY");
  // Double-check for optimized version
  if(d_lastValid.isNull()) return Proof();
  return d_lastValid.getProof();
}


const Assumptions
SearchEngine::getAssumptions()
{
  if(!d_vcl->getTM()->withAssumptions())
    throw EvalException
      ("DUMP_ASSUMPTIONS cannot be used without assumptions activated");
  if(d_lastValid.isNull())
    throw EvalException
      ("DUMP_ASSUMPTIONS must be called only after successful QUERY");
  // Double-check for optimized version
  if(d_lastValid.isNull()) return Assumptions();
  return d_lastValid.getAssumptions();
}

/*! 
 * Given a proof of FALSE ('res') from an assumption 'e', derive the
 * proof of the inverse of e.
 *
 * \param res is a proof of FALSE
 * \param e is the assumption used in the above proof
 */
bool SearchEngine::processResult(const Theorem& res, const Expr& e)
{
  // Result must be either Null (== SAT) or false (== unSAT)
  DebugAssert(res.isNull() || res.getExpr().isFalse(),
	      "processResult: bad input"
	      + res.toString());
  if(res.isNull()) {
    // Didn't prove valid (if CVC-lite is complete, it means invalid).
    // The assumptions for the counterexample must have been already
    // set by checkSAT().
    d_lastValid = Theorem(); // Reset last proof
    // Remove !e, keep just the counterexample
    d_lastCounterExample.erase(!e);
    if (e.isNot()) d_lastCounterExample.erase(e[0]);
    return false;
  } else {
    // Proved valid
    Theorem res2 = d_rules->eliminateSkolemAxioms(res, d_query_axioms);
    if(e.isNot())
      d_lastValid = d_rules->negIntro(e, res2);
    else
      d_lastValid = d_rules->proofByContradiction(e, res2);
    d_lastCounterExample.clear(); // Reset counterexample
    return true;
  }
}


Theorem SearchEngine::checkValidThm(const Expr& e) {
  d_query_axioms.clear();
  if(checkValid(e)) return d_lastValid;
  else return Theorem();
}


Theorem SearchEngine::restartThm(const Expr& e) {
  if (restart(e)) return d_lastValid;
  else return Theorem();
}


void 
SearchEngine::enqueueCNF(const Theorem& beta) {
  TRACE("mycnf", "enqueueCNF(", beta, ") {");
  if(*d_origFormulaOption)
    addCNFFact(beta);

  enqueueCNFrec(beta);
  TRACE_MSG("mycnf", "enqueueCNF => }");
}


void 
SearchEngine::enqueueCNFrec(const Theorem& beta) {
  Theorem theta = beta;

  TRACE("mycnf","enqueueCNFrec(", theta.getExpr(), ") { ");
  TRACE("cnf-clauses", "enqueueCNF call", theta.getExpr(), "");
  // The theorem scope shouldn't be higher than current
  DebugAssert(theta.getScope() <= d_vcl->scopeLevel(),
	      "\n theta.getScope()="+int2string(theta.getScope())
	      +"\n scopeLevel="+int2string(d_vcl->scopeLevel())
	      +"\n e = "+theta.getExpr().toString());

  Expr thetaExpr = theta.getExpr();

  if(d_enqueueCNFCache.count(thetaExpr) > 0) {
    TRACE_MSG("mycnf", "enqueueCNFrec[cached] => }");
    return;
  }

  d_enqueueCNFCache[thetaExpr] = true;

  // Strip double negations first
  while(thetaExpr.isNot() && thetaExpr[0].isNot()) {
    theta = d_core->notNotElim(theta);
    thetaExpr = theta.getExpr();
    // Cache all the derived theorems too
    d_enqueueCNFCache[thetaExpr] = true;
  }

  // Check for a propositional literal
  if(isPropLiteral(thetaExpr)) {
    theta = d_core->iffMP(theta, replaceITE(thetaExpr));
    DebugAssert(!*d_ifLiftOption || isLiteral(theta.getExpr()),
		"Must be a literal: theta = "
		+theta.getExpr().toString());
    addCNFFact(theta);
    TRACE_MSG("mycnf", "enqueueCNFrec[literal] => }");
    TRACE("cnf-clauses", "literal with proofs(", theta.getExpr(), ")");
    return;
  }

  // Obvious optimizations for AND and OR  
  switch(thetaExpr.getKind()) {
  case AND:
    // Split up the top-level conjunction and translate individually
    for(int i=0; i<thetaExpr.arity(); i++)
      enqueueCNFrec(d_core->andElim(theta, i));
    TRACE_MSG("mycnf", "enqueueCNFrec[AND] => }");
    return;
  case OR: {
    // Check if we are already in CNF (ignoring ITEs in the terms),
    // and if we are, translate term ITEs on the way
    bool cnf(true);
    for(int c=0, cend=thetaExpr.arity(); c<cend && cnf; ++c) {
      if(!isPropLiteral(thetaExpr[c]))
	cnf = false;
    }
    if(cnf) {
      vector<Theorem> thms;
      vector<unsigned int> changed;
      for(int c=0, cend=thetaExpr.arity(); c<cend; ++c) {
	Theorem thm = replaceITE(thetaExpr[c]);
	if(thm.getLHS() != thm.getRHS()) {
	  DebugAssert(!*d_ifLiftOption || isLiteral(thm.getRHS()),
		      "thm = "+thm.getExpr().toString());
	  thms.push_back(thm);
	  changed.push_back(c);
	}
      }
      if(changed.size() > 0) {
	Theorem rewrite =
	  d_core->substitutivityRule(theta.getExpr(), changed, thms);
	theta = d_core->iffMP(theta, rewrite);
      }
      DebugAssert(!*d_ifLiftOption || isClause(theta.getExpr()),
		  "theta = "+theta.getExpr().toString());
      addCNFFact(theta);
      TRACE_MSG("mycnf", "enqueueCNFrec[cnf] => }");
      return; 
    }
    break;
  }
  case IFF: { // Check for "var <=> phi" and "phi <=> var"
    const Expr& t0 = thetaExpr[0];
    const Expr& t1 = thetaExpr[1];
     if(isPropLiteral(t1)) {
       if(!isLiteral(t1))
	 theta = d_core->transitivityRule(theta, replaceITE(t1));
      applyCNFRules(theta);
      return;
    } else if(isPropLiteral(thetaExpr[0])) {
      theta = d_core->symmetryRule(theta);
      if(!isLiteral(t0))
	theta = d_core->transitivityRule(theta, replaceITE(t0));
      applyCNFRules(theta);
      return;
    }
    break;
  }
  case ITE: {
    if(isPropLiteral(thetaExpr[0]) && isPropLiteral(thetaExpr[1])
       && isPropLiteral(thetaExpr[2])) {
      // Replace ITEs
      vector<Theorem> thms;
      vector<unsigned int> changed;
      for(int c=0, cend=thetaExpr.arity(); c<cend; ++c) {
	Theorem thm = replaceITE(thetaExpr[c]);
	if(thm.getLHS() != thm.getRHS()) {
	  DebugAssert(!*d_ifLiftOption || isLiteral(thm.getRHS()),
		      "thm = "+thm.getExpr().toString());
	  thms.push_back(thm);
	  changed.push_back(c);
	}
      }
      if(changed.size() > 0) {
	Theorem rewrite =
	  d_core->substitutivityRule(theta.getExpr(), changed, thms);
	theta = d_core->iffMP(theta, rewrite);
      }
      // Convert to clauses
      Theorem thm = d_core->iteToClauses(theta);
      DebugAssert(thm.getExpr().isAnd() && thm.getExpr().arity()==2,
		  "enqueueCNFrec [ITE over literals]\n thm = "
		  +thm.getExpr().toString());
      addCNFFact(d_core->andElim(thm, 0));
      addCNFFact(d_core->andElim(thm, 1));
      return;
    }
    break;
  }
  default: 
    break;
  }

  // Now do the real work
  Theorem res = findInCNFCache(theta.getExpr());
  if(!res.isNull()) {
    Theorem thm = d_core->iffMP(theta, res);
    DebugAssert(isLiteral(thm.getExpr()), "thm = "+thm.getExpr().toString());
    addCNFFact(thm);
    TRACE("cnf-clauses", "enqueueCNFrec call[cache hit]:(",
	  thm.getExpr(), ")");
    applyCNFRules(res);
    TRACE_MSG("mycnf", "enqueueCNFrec[cached] => }");
    return;
  }

  //  std::vector<Theorem> clauses;
  Theorem result;
  // (\phi <==> v)
  result = varIntroSkolem(theta.getExpr(), false);

  TRACE("mycnf", "enqueueCNFrec: varIntroSkolem => ", result.getExpr(),
	" @ "+int2string(result.getScope())
	+" (current="+int2string(scopeLevel())+")");

  // result = skolemize(result, false);

  TRACE("mycnf", "enqueueCNFrec: skolemize => ", result.getExpr(),
	" @ "+int2string(result.getScope())
	+" (current="+int2string(scopeLevel())+")");
  DebugAssert(result.isRewrite(),
	      "SearchEngine::CNF(): result = "+result.toString());
  DebugAssert(!isPropLiteral(result.getLHS()) && 
	      isPropLiteral(result.getRHS()),
	      "SearchEngine::CNF(): result = "+result.toString());
  DebugAssert(result.getLHS() == theta.getExpr(),
	      "SearchEngine::CNF(): result = "+result.toString()
	      +"\n theta = "+theta.toString());
  
  //enqueue v
  Theorem var(d_core->iffMP(theta, result));
  // Add variable to the set of CNF vars
  d_cnfVars[var.getExpr()] = true;
  TRACE("mycnf", "enqueueCNFrec: theta = ", theta.getExpr(),
	" @ "+int2string(theta.getScope())
	+" (current="+int2string(scopeLevel())+")");
  TRACE("mycnf", "enqueueCNFrec: var = ", var.getExpr(),
	" @ "+int2string(var.getScope())
	+" (current="+int2string(scopeLevel())+")");
  DebugAssert(isLiteral(var.getExpr()), "var = "+var.getExpr().toString());
  addCNFFact(var);
  // phi <=> v
  addToCNFCache(result);
  applyCNFRules(result);
  TRACE_MSG("mycnf", "enqueueCNFrec => }");
}


/*!
 * \param thm is the input of the form phi <=> v
 */
void
SearchEngine::applyCNFRules(const Theorem& thm) {
  TRACE("mycnf", "applyCNFRules(", thm.getExpr(), ") { ");
  
  Theorem result = thm;
  DebugAssert(result.isRewrite(),
	      "SearchEngine::applyCNFRules: input must be an iff: " + 
	      result.getExpr().toString());
  Expr lhs = result.getLHS();
  Expr rhs = result.getRHS();
  DebugAssert(isLiteral(rhs),
	      "SearchEngine::applyCNFRules: rhs of input must be literal: " +
	      result.getExpr().toString());

  // Eliminate negation first
  while(result.getLHS().isNot())
    result = d_core->iffContrapositive(result);
  lhs = result.getLHS();
  rhs = result.getRHS();
  
  CDMap<Expr,bool>::iterator it = d_applyCNFRulesCache.find(lhs);
  if(it == d_applyCNFRulesCache.end())
    d_applyCNFRulesCache[lhs] = true;
  else {
    TRACE_MSG("mycnf","applyCNFRules[temp cache] => }");
    return;
  }
  
  // Catch the trivial case v1 <=> v2
  if(isPropLiteral(lhs)) {
    if(!isLiteral(lhs)) {
      Theorem replaced = d_core->symmetryRule(replaceITE(lhs));
      result = d_core->transitivityRule(replaced, result);
      lhs = result.getLHS();
      DebugAssert(rhs == result.getRHS(),
		  "applyCNFRules [literal lhs]\n result = "
		  +result.getExpr().toString()
		  +"\n rhs = "+rhs.toString());
    }
    Theorem thm = d_core->iffToClauses(result);
    DebugAssert(thm.getExpr().isAnd() && thm.getExpr().arity()==2,
		"applyCNFRules [literal lhs]\n thm = "
		+thm.getExpr().toString());
    addCNFFact(d_core->andElim(thm, 0));
    addCNFFact(d_core->andElim(thm, 1));
    return;
  }
  
  // Check the kids in e[0], replace them with cache[e[0][i]], rebuild thm
  vector<unsigned> changed;
  vector<Theorem> substitutions;
  int c=0;
  for(Expr::iterator j = lhs.begin(), jend = lhs.end(); j!=jend; ++c,++j) {
    const Expr& phi = *j;
    if(!isPropLiteral(phi)) {
      Theorem phiIffVar = findInCNFCache(phi);
      if(phiIffVar.isNull()) {
	// Create a new variable for this child
	phiIffVar = varIntroSkolem(phi, false);
	addToCNFCache(phiIffVar);
      }
      DebugAssert(phiIffVar.isRewrite(),
		  "SearchEngine::applyCNFRules: result = " +
		  result.toString());
      DebugAssert(phi == phiIffVar.getLHS(),
		  "SearchEngine::applyCNFRules:\n phi = " + 
		  phi.toString()
		  + "\n\n phiIffVar = " + phiIffVar.toString());
      DebugAssert(isLiteral(phiIffVar.getRHS()),
		  "SearchEngine::applyCNFRules: phiIffVar = " +
		  phiIffVar.toString());
      changed.push_back(c);
      substitutions.push_back(phiIffVar);
      applyCNFRules(phiIffVar);
    }
  }
  if(changed.size() > 0) {
    Theorem subst = 
      d_core->substitutivityRule(lhs, changed, substitutions);
    subst = d_core->symmetryRule(subst);
    result = d_core->transitivityRule(subst, result);
  }

  switch(result.getLHS().getKind()) {
  case AND:
    result = d_core->andCNFRule(result);
    break;
  case OR:
    result = d_core->orCNFRule(result);
    break;
  case IFF:
    result = d_core->iffCNFRule(result);
    break;
  case IMPLIES:
    result = d_core->impCNFRule(result);
    break;
  case ITE:
    result = d_core->iteCNFRule(result);
    break;
  default:
    DebugAssert(false,
		"SearchEngine::applyCNFRules: "
		"the input operator must be and|or|iff|imp|ite\n result = " + 
		result.getLHS().toString());
    break;
  }

  // FIXME: eliminate this once debugged
  Theorem clauses(result);

  /*
  if(EXISTS == result.getExpr().getKind()) {
    result = skolemize(result, false);

    DebugAssert(result.getExpr().isAnd() && 2 == result.getExpr().arity(),
		"SearchEngine::applyCNFRules:"
		"the result must be an *and theorem*:" + 
		result.toString());

    clauses = d_core->andElim(result,0); // CNF[v <=> OP(v1, ..., vn)]
    result = d_core->andElim(result, 1);
    const Expr& ExprIffnewVar = result.getExpr();
    DebugAssert(ExprIffnewVar.isAnd() || ExprIffnewVar.isIff(),
		"SearchEngine::applyCNFRules:"
		"we need a big conjunction here or a single iff" + 
		ExprIffnewVar.toString());
    
    if(ExprIffnewVar.isAnd()) {
      vector<Theorem> res;
      for(int count = 0; count < ExprIffnewVar.arity(); ++count) {
	res.push_back(d_core->andElim(result, count));
	DebugAssert(res.back().isRewrite() && 
		    !isPropLiteral(res.back().getLHS()), 
		    "SearchEngine::applyCNFRules: "
		    "res must be an Iff between new var and boolean expr: "+
		    res.back().toString());
	DebugAssert(isLiteral(res.back().getRHS()),
		    "SearchEngine::applyCNFRules: "
		    "rhs of res must be a literal: " + res.back().toString());
	addToCNFCache(res.back());
      }
      DebugAssert(((int)res.size()) == ExprIffnewVar.arity(),
		  "SearchEngine::applyCNFRules: res.size() = "
		  +int2string(res.size())+", arity = "
		  +int2string(ExprIffnewVar.arity()));
      for(size_t i = 0; i < res.size(); ++i) {
	DebugAssert(res[i].isRewrite() && 
		    !isPropLiteral(res[i].getLHS()), 
		    "SearchEngine::applyCNFRules:"
		    "res must be an Iff between new var and boolean expr"+
		    res[i].toString());
	DebugAssert(isLiteral(res[i].getRHS()),
		    "SearchEngine::applyCNFRules:"
		    "rhs of res must be a literal" + res[i].toString());
	applyCNFRules(res[i]);
      }
    } else {
      DebugAssert(result.isRewrite() && 
		  !isPropLiteral(result.getLHS()) &&
		  isPropLiteral(result.getRHS()),		 
		  "SearchEngine::applyCNFRules:"
		  "rhs of res must be a literal" + result.toString());
      addToCNFCache(result);
      applyCNFRules(result);
    }
  } else {
    // We are already in clausal normal form.  Split up the clauses
    // and enqueue them individually.
    clauses = result;
  }
  */

  // Enqueue the clauses
  DebugAssert(!clauses.isNull(), "Oops!..");
  DebugAssert(clauses.getExpr().isAnd(), "clauses = "
	      +clauses.getExpr().toString());

  // The clauses may containt term ITEs, and we need to translate those
  
  for(int i=0, iend=clauses.getExpr().arity(); i<iend; ++i) {
    Theorem clause(d_core->andElim(clauses,i));
    TRACE("cnf-clauses", "clauses["+int2string(i)+"] = ",clause.getExpr(),"");

    // Go over the literals, replace ITEs and/or bitblast atomic formulas
    const Expr& clauseExpr = clause.getExpr();
    vector<Theorem> substThms;
    vector<unsigned> changed;
    for(int j=0, jend=clauseExpr.arity(); j<jend; ++j) {
      Theorem thm = replaceITE(clauseExpr[j]);
      if(thm.getLHS() != thm.getRHS()) {
	substThms.push_back(thm);
	changed.push_back(j);
      }
    }
    if(changed.size() > 0) {
      Theorem subst(d_core->substitutivityRule
		    (clauseExpr,changed,substThms));
      clause = d_core->iffMP(clause, subst);
    }
    DebugAssert(isPropClause(clause.getExpr()),
		"clause = "+clause.getExpr().toString());
    addCNFFact(clause);
  }
  TRACE_MSG("mycnf","applyCNFRules => }");
}


bool SearchEngine::isLiteral(const Expr& e) {
  return d_core->isLiteral(e);
}


bool SearchEngine::isClause(const Expr& e) {
  if(isLiteral(e)) return true;
  if(!e.isOr()) return false;

  bool cnf(true);
  for(Expr::iterator i=e.begin(), iend=e.end(); i!=iend && cnf; ++i)
    cnf = isLiteral(*i);
  return cnf;
}


bool
SearchEngine::isPropAtom(const Expr& e) {
  DebugAssert(d_core->getType(e).isBool(), "SearchEngine::isPropAtom: e = "
	      +e.toString()+" : "+d_core->getType(e).toString());
  switch(e.getKind()) {
  case AND:
  case OR:
  case IFF:
  case IMPLIES:
  case ITE:
  case NOT:
    return false;
  default:
    return true;
  }
}

bool
SearchEngine::isPropLiteral(const Expr& e) {
  switch(e.getKind()) {
  case NOT:
    return isPropAtom(e[0]);
  default:
    return isPropAtom(e);
  }
}


bool
SearchEngine::isPropClause(const Expr& e) {
  if(isPropLiteral(e)) return true;
  if(!e.isOr()) return false;

  bool cnf(true);
  for(Expr::iterator i=e.begin(), iend=e.end(); i!=iend && cnf; ++i)
    cnf = isPropLiteral(*i);
  return cnf;
}


bool
SearchEngine::isGoodSplitter(const Expr& e) {
  if(!*d_ignoreCnfVarsOption) return true;
  TRACE("isGoodSplitter", "isGoodSplitter(", e, ") {");
  // Check for var being an auxiliary CNF variable
  const Expr& var = e.isNot()? e[0] : e;
  bool res(!isCNFVar(var));
  TRACE("isGoodSplitter", "isGoodSplitter => ", res? "true" : "false", " }");
  return res;
}


void
SearchEngine::addToCNFCache(const Theorem& thm) {
  TRACE("mycnf", "addToCNFCache(", thm.getExpr(), ")");

  d_vcl->getStatistics().counter("CNF New Vars")++;

  Theorem result = thm;
  DebugAssert(result.isRewrite(),
	      "SearchEngine::addToCNFCache: input must be an iff: " + 
	      result.getExpr().toString());
  // Record the CNF variable
  d_cnfVars[thm.getRHS()] = true;

  Theorem t(thm);
  Expr phi = thm.getLHS();
  while(phi.isNot()) {
    t = d_core->iffContrapositive(thm);
    phi = phi[0];
  }
  DebugAssert(d_cnfCache.count(phi) == 0, 
	      "thm = "+thm.getExpr().toString() + 
	      "\n t = "+ t.getExpr().toString());
  //d_cnfCache.insert(phi, t);
  d_cnfCache.insert(phi, t, d_bottomScope);
}


Theorem 
SearchEngine::findInCNFCache(const Expr& e) { 
  TRACE("mycnf", "findInCNFCache(", e, ") { ");
  Expr phi(e);

  int numNegs(0);
  // Strip all the top-level negations from e
  while(phi.isNot()) {
    phi = phi[0]; numNegs++;
  }
  CDMap<Expr,Theorem>::iterator it = d_cnfCache.find(phi);
  if(it != d_cnfCache.end()) {
    // IF_DEBUG(debugger.counter("CNF cache hits")++);
    d_vcl->getStatistics().counter("CNF cache hits")++;
    Theorem thm = (*it).second;
    
    DebugAssert(thm.isRewrite() && thm.getLHS() == phi,
		"SearchEngine::findInCNFCache: thm must be an iff: " + 
		thm.getExpr().toString());
//     DebugAssert(!isPropLiteral(thm.getLHS()) &&
// 		isPropLiteral(thm.getRHS()),
// 		"SearchEngine::findInCNFCache: "
// 		"input expression cannot be true or false: " + 
// 		thm.getExpr().toString());
//     DebugAssert(!isPropLiteral(phi) &&
// 		phi == thm.getLHS(),
// 		"SearchEngine::findInCNFCache: "
// 		"input expression cannot be prop literal: " + 
// 		phi.toString());
    
    // Now put all the negations back.  If the number of negations is
    // odd, first transform phi<=>v into !phi<=>!v.  Then apply
    // !!phi<=>phi rewrite as many times as needed.
    if(numNegs % 2 != 0) {
      thm = d_core->iffContrapositive(thm); 
      numNegs--;
    }
    for(; numNegs > 0; numNegs-=2) {
      Theorem t = d_core->rewriteNotNot(!!(thm.getLHS()));
      thm = d_core->transitivityRule(t,thm);
    }

    DebugAssert(numNegs == 0, "numNegs = "+int2string(numNegs));
    TRACE("mycnf", "findInCNFCache => ", thm.getExpr(), " }");
    return thm;
  }
  TRACE_MSG("mycnf", "findInCNFCache => null  }");
  return Theorem();
}

/*!
 * Strategy:
 *
 * For a given atomic formula phi(ite(c, t1, t2)) which depends on a
 * term ITE, generate an equisatisfiable formula:
 *
 * phi(x) & ite(c, t1=x, t2=x),
 *
 * where x is a new variable.
 *
 * The phi(x) part will be generated by the caller, and our job is to
 * assert the 'ite(c, t1=x, t2=x)', and return a rewrite theorem
 * phi(ite(c, t1, t2)) <=> phi(x).
 */
Theorem
SearchEngine::replaceITE(const Expr& e) {
  TRACE("replaceITE","replaceITE(", e, ") { ");
  Theorem res;

  CDMap<Expr,Theorem>::iterator i=d_replaceITECache.find(e),
    iend=d_replaceITECache.end();
  if(i!=iend) {
    TRACE("replaceITE", "replaceITE[cached] => ", (*i).second, " }");
    return (*i).second;
  }

  if(*d_ifLiftOption) {
    DebugAssert(isPropLiteral(e),
		"SearchEngine::replaceITE: "
		"input must be a propositional literal: e = " + e.toString());

    if(isLiteral(e))
      res = d_core->reflexivityRule(e);
    else if(e.isNot()) {
      res = replaceITErec(e[0]);
      res = d_core->iffContrapositive(res);
    }
    else
      res = replaceITErec(e);
  } else
    res = d_core->reflexivityRule(e);

  if(isLiteral(res.getRHS()))
    res = d_core->transitivityRule(res, d_core->rewriteLiteral(res.getRHS()));
  // If 'res' is e<=>phi, and the resulting formula phi is not a
  // literal, introduce a new variable x, enqueue phi<=>x, and return
  // e<=>x.
  if(!isPropLiteral(res.getRHS())) {
    Theorem varThm(findInCNFCache(res.getRHS()));
    if(varThm.isNull()) {
      varThm = varIntroSkolem(res.getRHS(), false);
      addToCNFCache(varThm);
    }
    applyCNFRules(varThm);
    res = d_core->transitivityRule(res, varThm);
  }

  TRACE("replaceITE", "replaceITE => ", res, " }");
  d_replaceITECache[e] = res;
  return res;
}


Theorem
SearchEngine::replaceITErec(const Expr& e) {
  TRACE("replaceITE","replaceITErec(", e, ") { ");

  DebugAssert(isPropAtom(e) || d_core->isTerm(e), 
	      "SearchEngine::replaceITErec("+e.toString()+")");

  CDMap<Expr,Theorem>::iterator i=d_replaceITECache.find(e),
    iend=d_replaceITECache.end();
  if(i!=iend) {
    TRACE("replaceITE", "replaceITErec[cache] => ",
	  (*i).second.getExpr(), " }");
    return (*i).second;
  }

  // Check if e contains ITEs
  bool isTerm(d_core->isTerm(e));
  // Ignore atomic terms, but not atomic formulas, since we need to
  // bit-blast them through simplifier
  if((isTerm && d_core->isAtomic(e))
     /* || (!isTerm && d_core->isAtomicFormula(e)) */) {
    Theorem res= d_core->reflexivityRule(e);
    d_replaceITECache[e] = res;
    TRACE("replaceITE", "replaceITErec[no ITEs] => ", res, " }");
    return res;
  }
  
  DebugAssert(!e.isNot(),
	      "replaceITErec: e = " + e.toString());

  if(e.isITE()) {
    Theorem res;
    // If this is a Boolean ITE, punt.  We may have Boolean arguments
    // to functions, or Boolean record fields, etc., and we don't
    // translate them here.
    if(!d_core->isTerm(e)) {
      res= d_core->reflexivityRule(e);
      d_replaceITECache[e] = res;
      TRACE("replaceITE", "replaceITErec[BOOLEAN] => ", res, " }");
      return res;
    }

    res = findInCNFCache(e);
    if(res.isNull()) {
      // ite(cond,e1,e2)=x
      res = varIntroSkolem(e, false);
      addToCNFCache(res);
    }
    //thm: ite(cond,e1,e2)=x <==> ite(cond,e1=x,e2=x)
    Theorem thm = d_core->ifLiftRule(res.getExpr(), 0);
    //thm: ite(cond,e1=x,e2=x)
    thm = d_core->iffMP(res,thm);
    enqueueCNFrec(thm);
    
    d_replaceITECache[e] = res;
    TRACE("replaceITE", "replaceITErec[ITE] => ", res, " }");
    return res;
  }

  // e is not an ITE; recurse into the children
  vector<Theorem> substThms;
  vector<unsigned> changed;

  unsigned int c=0;
  for(Expr::iterator i=e.begin(),iend=e.end(); i!=iend; i++, c++) {
    Theorem thm = replaceITErec(*i);
    if(thm.getLHS() != thm.getRHS()) {
      substThms.push_back(thm);
      changed.push_back(c);
    }
  }

  Theorem res;

  if(changed.size() > 0)
    res = d_core->substitutivityRule(e,changed,substThms);
  else
    res = d_core->reflexivityRule(e);

  /*
  if(!d_core->isTerm(res.getRHS())) {
    DebugAssert(isLiteral(res.getRHS()), "res = "+res.getExpr().toString());
    // Rewrite atomic formulas
    res = d_core->transitivityRule(res, d_core->rewriteLiteral(res.getRHS()));
    // If 'res' is e<=>phi, and the resulting formula phi is not a
    // literal, introduce a new variable x, enqueue phi<=>x, and return
    // e<=>x.
    if(!isPropAtom(res.getRHS())) {
      Theorem varThm(findInCNFCache(res.getRHS()));
      if(varThm.isNull()) {
	varThm = d_core->varIntroRule(res.getRHS());
	varThm = skolemize(varThm, false);
	addToCNFCache(varThm);
      }
      enqueueCNFrec(varThm);
      res = d_core->transitivityRule(res, varThm);
    }
  }
  */

  TRACE("replaceITE","replaceITErec => ", res.getExpr(), " }");
  d_replaceITECache[e] = res;
  return res;
}


Expr SearchEngine::getImpliedLiteral()
{
  if (d_literals.size() == 0) return Expr();
  Expr res = d_literals.front();
  d_literals.pop_front();
  return res;
}
